package com.starbuds.server.api.common.extension;

import cn.hutool.core.codec.Base64;
import cn.hutool.crypto.Mode;
import cn.hutool.crypto.Padding;
import cn.hutool.crypto.SecureUtil;
import cn.hutool.crypto.asymmetric.KeyType;
import cn.hutool.crypto.asymmetric.RSA;
import cn.hutool.crypto.symmetric.AES;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.api.common.constant.HttpError;
import com.starbuds.server.api.common.inf.PermitGuest;
import com.starbuds.server.api.common.util.SpringContextUtil;
import com.starbuds.server.common.constant.common.ClientType;
import com.starbuds.server.common.constant.common.SignType;
import com.starbuds.server.common.pojo.api.HeaderPackage;
import com.starbuds.server.common.pojo.user.TokenBO;
import com.starbuds.server.common.service.user.SecurityService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.jboss.resteasy.core.ResourceMethodInvoker;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.security.DenyAll;
import javax.annotation.security.PermitAll;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.container.ContainerRequestContext;
import javax.ws.rs.container.ContainerRequestFilter;
import javax.ws.rs.core.Context;
import javax.ws.rs.ext.Provider;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Map;

/**
 * 接口访问安全拦截器
 *
 * @author Lay
 * @date 2019-02-18
 */
@Slf4j
@Component
@Provider
public class SecurityInterceptor implements ContainerRequestFilter {

    private static final String AES_KEY = "*TiMi#Chat@2020*";

    @Reference(version = "1.0.0")
    SecurityService securityService;

    @Value("${api.privateKey}")
    private String apiPrivateKey;

    @Context
    HttpServletRequest request;

    static final long SIGN_TIMEOUT = 60 * 1000L;

    @Override
    public void filter(ContainerRequestContext requestContext) throws IOException {
        String reqPath = requestContext.getUriInfo().getPath();

        // 打印请求日志
        String qs = paramsMapToQueryString(request.getParameterMap());
        String headers = getHeaderString();
        log.info(String.format("\n请求接口:%s\n请求方式:%s\n请求参数:%s\n请求头:%s", reqPath, requestContext.getMethod(), qs, headers));

        // 权限校验部分
        ResourceMethodInvoker methodInvoker = (ResourceMethodInvoker) requestContext.getProperty("org.jboss.resteasy.core.ResourceMethodInvoker");
        Method method = methodInvoker.getMethod();

        // 请求头解包
        HeaderPackage headerPackage = getHeaderPackage();

        // 请求头对象注入
        requestContext.setProperty("header", headerPackage);

        // 如果不是除了所有的权限
        if (!method.isAnnotationPresent(PermitAll.class)) {
            // 如果是拒绝所有访问
            if (method.isAnnotationPresent(DenyAll.class)) {
                requestContext.abortWith(HttpError.ACCESS_FORBIDDEN);
                return;
            }

            // 请求头判断
            if (headerPackage == null) {
                requestContext.abortWith(HttpError.ACCESS_DENIED);
                return;
            }

            // 请求头时间校验
            if (!SpringContextUtil.isDev() && headerPackage.getTimestamp() + SIGN_TIMEOUT < System.currentTimeMillis()) {
                log.error("签名已过期");
                requestContext.abortWith(HttpError.ACCESS_DENIED);
                return;
            }

            // 解析令牌
            TokenBO tokenBO = securityService.getToken(headerPackage.getToken());
            if (tokenBO != null && tokenBO.getUserId() > 0) {
                // 令牌注入上下文
                requestContext.setProperty("token", tokenBO);
            } else {
                if (!method.isAnnotationPresent(PermitGuest.class)) {
                    requestContext.abortWith(HttpError.TOKEN_INVALID);
                    return;
                }
            }
        }
    }

    private HeaderPackage getHeaderPackage() {
        String pkg = request.getHeader("pkg");

        if (StringUtils.isNotBlank(pkg)) {
            // 尝试解包
            try {
                boolean isAES = false;
                String tmpPkg = Base64.decodeStr(pkg);
                boolean isCorrect = tmpPkg.startsWith("{") && tmpPkg.endsWith("}");
                if (!isCorrect) {
                    tmpPkg = decodeAES(pkg);
                    if (StringUtils.isNotBlank(tmpPkg)) {
                        isAES = true;
                    }
                }
                log.info("decode pkg[AES {}]: {}", isAES, tmpPkg);
                HeaderPackage headerPackage = JSON.parseObject(tmpPkg, HeaderPackage.class);

                // 参数校验
                if (headerPackage.getClientType() == null ||
                        headerPackage.getAppType() == null ||
                        headerPackage.getPackageName() == null ||
                        headerPackage.getTimestamp() == null ||
                        headerPackage.getSignType() == null ||
                        headerPackage.getSign() == null ||
                        headerPackage.getToken() == null ||
                        headerPackage.getDeviceId() == null) {
                    log.error("请求头参数不完整 {} {} {} {} {} {} {} {}", headerPackage.getClientType(), headerPackage.getAppType(), headerPackage.getPackageName(), headerPackage.getTimestamp(), headerPackage.getSignType(), headerPackage.getSign(), headerPackage.getToken(), headerPackage.getDeviceId());
                    return null;
                }

                // 芝麻渠道强制 RSA
                boolean isClient = headerPackage.getClientType() == ClientType.Android || headerPackage.getClientType() == ClientType.iOS;
                Long zhimaChannel = 78139153711104L;
                if (isClient && zhimaChannel.equals(headerPackage.getChannelId()) && !isAES) {
                    log.warn("该渠道强制AES");
                    return null;
                }

                // 签名校验
                if (verifySign(headerPackage)) {
                    return headerPackage;
                }
                log.error("签名校验不通过");
            } catch (Exception e) {
                log.error("请求头解包失败", e);
            }
        }

        return null;
    }

    private String decodeRSA(String pkg) {
        RSA rsa = new RSA(apiPrivateKey, null);
        return rsa.decryptStr(pkg, KeyType.PrivateKey);
    }

    private String decodeAES(String pkg) {
        AES aes = new AES(Mode.ECB, Padding.NoPadding, AES_KEY.getBytes());
        return aes.decryptStr(pkg);
    }

    /**
     * 签名校验
     *
     * @return
     */
    private boolean verifySign(HeaderPackage headerPackage) {
        String toSign = headerPackage.toQueryString();
        if (ClientType.iOS == headerPackage.getClientType() || ClientType.Android == headerPackage.getClientType()) {
            toSign = String.format("%s%s.%s.security", toSign, headerPackage.getPackageName(), headerPackage.getClientType().getName());
        } else {
            toSign = String.format("%s%s.security", toSign, headerPackage.getPackageName());
        }
        String sign = "";
        switch (headerPackage.getSignType()) {
            case SignType.MD5:
                sign = SecureUtil.md5(toSign);
                break;
            case SignType.SHA1:
                sign = SecureUtil.sha1(toSign);
                break;
            case SignType.SHA256:
                sign = SecureUtil.sha256(toSign);
                break;
            default:
                break;
        }

        return StringUtils.isNotBlank(sign) && sign.equalsIgnoreCase(headerPackage.getSign());
    }

    private String paramsMapToQueryString(Map<String, String[]> map) {
        if (map == null) {
            return "";
        }

        StringBuilder sbQuery = new StringBuilder();
        for (String key : map.keySet()) {
            for (String val : map.get(key)) {
                sbQuery.append(String.format("&%s=%s", key, val));
            }
        }
        if (sbQuery.length() > 0) {
            return sbQuery.toString().substring(1);
        }

        return "";
    }

    private String getHeaderString() {
        Enumeration HeaderKey = request.getHeaderNames();
        StringBuilder sbHeader = new StringBuilder();
        while (null != HeaderKey && HeaderKey.hasMoreElements()) {
            String headerKey = (String) HeaderKey.nextElement();
            sbHeader.append(String.format("%s=%s&", headerKey, request.getHeader(headerKey)));
        }
        return sbHeader.toString();
    }
}
